using System.Security.Claims;
using FitConnect.Models.v1.Api;
using Microsoft.IdentityModel.JsonWebTokens;
using Microsoft.IdentityModel.Tokens;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;

namespace FitConnect.Models;

public enum EventType {
    NotSet,
    Create,
    Submit,
    Notify,
    Forward,
    Reject,
    Accept,
    Delete
}

public class Events {
    [JsonProperty("https://schema.fitko.de/fit-connect/events/submit-submission",
        NullValueHandling = NullValueHandling.Ignore)]
    public SubmitSubmissionEvent? SubmitSubmissionEvent { get; set; }

    [JsonProperty("https://schema.fitko.de/fit-connect/events/reject-submission",
        NullValueHandling = NullValueHandling.Ignore)]
    public RejectSubmissionEvent? RejectSubmissionEvent { get; set; }

    [JsonProperty("https://schema.fitko.de/fit-connect/events/create-submission",
        NullValueHandling = NullValueHandling.Ignore)]
    public CreatedSubmissionEvent? CreatedSubmissionEvent { get; set; }

    [JsonProperty("https://schema.fitko.de/fit-connect/events/notify-submission",
        NullValueHandling = NullValueHandling.Ignore)]
    public NotifySubmissionEvent? NotifySubmissionEvent { get; set; }

    [JsonProperty("https://schema.fitko.de/fit-connect/events/accept-submission",
        NullValueHandling = NullValueHandling.Ignore)]
    public AcceptSubmissionEvent? AcceptSubmissionEvent { get; set; }

    [JsonProperty("https://schema.fitko.de/fit-connect/events/delete-submission",
        NullValueHandling = NullValueHandling.Ignore)]
    public DeleteSubmissionEvent? DeleteSubmissionEvent { get; set; }
}

public interface ISubmissionEvent {
//     public abstract string EventSchema { get; }
//
// public string ToJson() {
//         return JsonConvert.SerializeObject(this);
//     }
}

public class NotifySubmissionEvent : ISubmissionEvent {
    [JsonProperty("notifyType", NullValueHandling = NullValueHandling.Ignore)]
    public string? NotifyType { get; set; }
}

public class AuthenticationTags {
    [JsonProperty("metadata", NullValueHandling = NullValueHandling.Ignore)]
    public string? Metadata { get; set; }

    [JsonProperty("data", NullValueHandling = NullValueHandling.Ignore)]
    public string? Data { get; set; }

    [JsonProperty("attachments", NullValueHandling = NullValueHandling.Ignore)]
    public Dictionary<string, string>? Attachments { get; set; }
}

public class CreatedSubmissionEvent : ISubmissionEvent {
}

public class AcceptSubmissionEvent : SubmitSubmissionEvent, ISubmissionEvent {
    public AcceptSubmissionEvent() {
    }

    public AcceptSubmissionEvent(AuthenticationTags payload) {
        this.AuthenticationTags = payload;
    }

    [JsonProperty("problems", NullValueHandling = NullValueHandling.Ignore)]
    public List<Problems>? Problems { get; set; }
}

public class DeleteSubmissionEvent : ISubmissionEvent {
}

public class SubmitSubmissionEvent : ISubmissionEvent {
    [JsonProperty("authenticationTags", NullValueHandling = NullValueHandling.Ignore)]
    public AuthenticationTags? AuthenticationTags { get; set; }
}

public class RejectSubmissionEvent : ISubmissionEvent {
    public RejectSubmissionEvent() {
        Problems = new List<Problems>();
    }

    public RejectSubmissionEvent(Problems[] problemsArray) {
        this.Problems = problemsArray.ToList();
    }

    [JsonProperty("problems", NullValueHandling = NullValueHandling.Ignore)]
    public List<Problems> Problems { get; set; }
}

public class SecurityEventToken {
    private readonly string? _submissionId;

    public const string CreateSubmissionSchema =
        "https://schema.fitko.de/fit-connect/events/create-submission";

    public const string SubmitSubmissionSchema =
        "https://schema.fitko.de/fit-connect/events/submit-submission";

    public const string NotifySubmissionSchema =
        "https://schema.fitko.de/fit-connect/events/notify-submission";

    public const string ForwardSubmissionSchema =
        "https://schema.fitko.de/fit-connect/events/forward-submission";

    public const string RejectSubmissionSchema =
        "https://schema.fitko.de/fit-connect/events/reject-submission";

    public const string AcceptSubmissionSchema =
        "https://schema.fitko.de/fit-connect/events/accept-submission";

    public const string DeleteSubmissionSchema =
        "https://schema.fitko.de/fit-connect/events/delete-submission";

    private SecurityEventToken(string jwtEncodedString) {
        TokenString = jwtEncodedString;
        EventType = DecodeEventType(Token.Claims);

        if (Token.Claims.All(c => c.Type != "iat"))
            return;

        var eventClaim = Token.Claims.Single(c => c.Type == "events").Value;
        Events = JsonConvert.DeserializeObject<Events>(eventClaim);

        var iat = Token.Claims.FirstOrDefault(c => c.Type == "iat")!.Value;
        var payloadData =
            JsonConvert.DeserializeObject<dynamic>(Base64UrlEncoder.Decode(Token.EncodedPayload));
        CompletePayload = payloadData;
        SubmissionId = payloadData!.sub.ToString().Split(':')[1];

        if (long.TryParse(iat, out var timeEpoch))
            EventTime = DateTime.UnixEpoch.AddSeconds(timeEpoch);
    }

    internal SecurityEventToken() {
    }

    public DateTime EventTime { get; set; }
    public EventType EventType { get; set; }
    public List<Problems>? Problems => Events?.RejectSubmissionEvent?.Problems;
    public Events? Events { get; set; }

    [JsonIgnore] public object? Payload { get; set; }
    public dynamic? CompletePayload { get; }
    public JsonWebToken Token => new(TokenString);
    public string? TokenString { get; set; }


    [JsonProperty("sub")]
    public string? SubmissionId {
        get => _submissionId?.Contains("submission:") ?? false
            ? _submissionId["submission:".Length..]
            : _submissionId;
        init => _submissionId = value;
    }

    public static SecurityEventToken FromJson(string json) {
        var result = JsonConvert.DeserializeObject<SecurityEventToken>(json) ??
                     throw new ArgumentException("Not a valid SET json");

        result.EventType = EventType.NotSet;
        if (result.Events?.SubmitSubmissionEvent != null) result.EventType = EventType.Submit;
        if (result.Events?.RejectSubmissionEvent != null) result.EventType = EventType.Reject;
        if (result.Events?.CreatedSubmissionEvent != null) result.EventType = EventType.Create;
        if (result.Events?.NotifySubmissionEvent != null) result.EventType = EventType.Notify;
        if (result.Events?.AcceptSubmissionEvent != null) result.EventType = EventType.Accept;
        if (result.Events?.DeleteSubmissionEvent != null) result.EventType = EventType.Delete;

        return result;
    }


    public static SecurityEventToken FromJwtEncodedString(string jwt) {
        var token = new JsonWebToken(jwt);
        var json = Base64UrlEncoder.Decode(token.EncodedPayload);
        var result = SecurityEventToken.FromJson(json);
        return result;
    }

    private EventType DecodeEventType(IEnumerable<Claim> claims) {
        var eventsClaim = claims.FirstOrDefault(c => c.Type == "events");
        if (eventsClaim == null) return EventType.NotSet;

        var events = JsonConvert.DeserializeObject<Dictionary<string, dynamic>>(eventsClaim.Value);
        Payload = JsonConvert
            .DeserializeObject<dynamic>(
                events?.Values?.FirstOrDefault()?.ToString());

        var contentData = JsonConvert.DeserializeObject(eventsClaim.Value);

        if (eventsClaim.Value.Contains(
                CreateSubmissionSchema))
            return EventType.Create;

        if (eventsClaim.Value.Contains(
                SubmitSubmissionSchema))
            return EventType.Submit;

        if (eventsClaim.Value.Contains(
                NotifySubmissionSchema))
            return EventType.Notify;
        if (eventsClaim.Value.Contains(
                ForwardSubmissionSchema))
            return EventType.Forward;
        if (eventsClaim.Value.Contains(
                RejectSubmissionSchema)) {
            //Problems = GetProblems(events?.Values?.FirstOrDefault()?.ToString() ?? "");
            return EventType.Reject;
        }


        if (eventsClaim.Value.Contains(AcceptSubmissionSchema))

            return EventType.Accept;


        if (eventsClaim.Value.Contains(
                DeleteSubmissionSchema))
            return EventType.Delete;

        return EventType.NotSet;
    }

    private static List<Problems> GetProblems(string payload) {
        if (payload == "") return new List<Problems>();

        var result = (JsonConvert.DeserializeObject<dynamic>(payload)?.problems as JArray)
            ?.Select(j => j as dynamic).ToList()
            ?.Select(p => new Problems {
                Title = p.title,
                Type = p.type,
                Detail = p.detail,
                Instance = p.instance
            })
            ?.ToList() ?? new List<Problems>();
        return result;
    }
}
